"""
Wrapper around the essential functionality  the `mbqc_scheduling crate`_.

.. _mbqc_scheduling crate:
   https://github.com/taeruh/mbqc_scheduling/blob/main/mbqc_scheduling
"""

from mbqc_scheduling._lib import SpacialGraph, PartialOrderGraph, Paths, Path, run
